"""
Example discussion module for asyncio-based subscribers.

This file is meant to demonstrate the different ways to consume data
from an `asyncio_for_robotics` subscriber, including:

- `wait_for_value`: get the latest value (can return the same value repeatedly).
- `wait_for_new`: wait for the most recent fresh message.
- `wait_for_next`: deterministically return the next message after the call.
- `listen`: async generator that streams new data but may skip intermediate messages.
- `listen_reliable`: async generator that queues all messages without skipping.

It also shows how to use timeouts to control asynchronous loops.
"""
import asyncio
from textwrap import dedent, fill
from typing import Any, Callable, Coroutine

from asyncio_for_robotics.core.sub import BaseSub, _MsgType
from asyncio_for_robotics.core.utils import soft_timeout


def brint(string: str):
    """print but creates a single paragraph of 70 columns"""
    print(fill(dedent(string)))


async def discuss(
    afor_sub: BaseSub[_MsgType],
    get_str_func: Callable[[_MsgType], str],
):

    sub = afor_sub
    print()
    brint(f"\nAwaiting value with `wait_for_value`")
    value = await sub.wait_for_value()
    brint(f"Received: {get_str_func(value)}")

    print()
    brint(
        f"""\nAwaiting value again with `wait_for_value` will return the same
        value as before."""
    )
    value = await sub.wait_for_value()
    brint(f"Received: {get_str_func(value)}")

    print()
    brint(
        f"""\nAwaiting NEW data with `wait_for_new`, will really wait for fresh
        data and return the most recent message."""
    )
    value = await sub.wait_for_new()
    brint(f"Received: {get_str_func(value)}")
    brint(f"`wait_for_value` is now: {get_str_func(await sub.wait_for_value())}")

    print()
    brint(
        f"""\n`wait_for_next` will also get fresh data. However will return a
        deterministic next-after-call value (first message after instantiation
                                             of the awaitable)"""
    )
    brint("Now, calling `wait_for_new` and `wait_for_next` simultaneously")
    new = sub.wait_for_new()
    next = sub.wait_for_next()
    brint(f"Value at time of call is: {get_str_func(await sub.wait_for_value())}")
    brint(f"Sleeping for 3 messages")
    for _ in range(3):
        await sub.wait_for_new()
    brint(f"wait_for_new: {get_str_func(await new)}")
    brint(f"wait_for_next: {get_str_func(await next)}")
    brint(
        f"""`wait_for_next` got exactly the next message after its call,
        whereas `wait_for_new` just displays the most recent value"""
    )

    print()
    brint(
        f"""\nAsync generators allow to continuously listen to data, simply in
        an async for loop. This can replace callbacks and handlers in
        ros/zenoh."""
    )
    count = 0
    async with soft_timeout(1):
        async for sample in sub.listen():
            brint(f"Processing: {get_str_func(sample)}")
            count += 1

    print()
    brint(
        f"""\nAs before with new/next, the `listen` method can miss messages.
        It only processes the latest message."""
    )
    count = 0
    async for sample in sub.listen():
        brint(f"Processing: {get_str_func(sample)}")
        brint(f"Sleeping for 0.5s")
        await asyncio.sleep(0.5)
        count += 1
        if count > 4:
            break

    print()
    brint(
        f"""\nOn the other hand, the `listen_reliable` method does not miss. It
        will properly build a queue of messages to process"""
    )
    count = 0
    async for sample in sub.listen_reliable(queue_size=40):
        brint(f"Processing: {get_str_func(sample)}")
        brint(f"Sleeping for 0.5s")
        await asyncio.sleep(0.5)
        count += 1
        if count > 4:
            break

    return



"""Implements a publisher and subscriber in the same node.

run with `python3 -m asyncio_for_robotics.example.ros2_pubsub`"""
import asyncio
from contextlib import suppress

import rclpy
from std_msgs.msg import String

from asyncio_for_robotics.core.utils import Rate
import asyncio_for_robotics.ros2 as afor

TOPIC = afor.TopicInfo(msg_type=String, topic="topic")


async def hello_world_publisher():
    # creates a standard ROS 2 publisher safely
    with afor.auto_session().lock() as node:
        pub = node.create_publisher(TOPIC.msg_type, TOPIC.topic, TOPIC.qos)
    start_time = None
    
    # Async for loop, itterating for every tick of a 2Hz timer
    async for t_ns in Rate(frequency=2).listen_reliable():  
        if start_time is None:
            start_time = t_ns
        payload = f"[Hello World! timestamp: {(t_ns-start_time)/1e9}s]"
        print(f"Publishing: {payload}")
        pub.publish(String(data=payload))  # sends payload (lock not necessary)


async def hello_world_subscriber():
    # creates an afor subscriber for a ROS 2 topic
    sub = afor.Sub(TOPIC.msg_type, TOPIC.topic, TOPIC.qos)
    
    # async for loop itterating for every messages received by the sub
    async for message in sub.listen_reliable():
        print(f"Received: {message.data}")


async def hello_world_pubsub():
    # starts both tasks.
    # Notice how easy it is to compose behaviors.
    sub_task = asyncio.create_task(hello_world_subscriber())
    pub_task = asyncio.create_task(hello_world_publisher())
    await asyncio.wait([pub_task, sub_task])


if __name__ == "__main__":
    rclpy.init()
    try:
        # suppress, just so we don't flood the terminal on exit
        with suppress(KeyboardInterrupt, asyncio.CancelledError):
            asyncio.run(hello_world_pubsub())  # starts asyncio executor
    finally:
        # cleanup. `finally` statment always executes
        afor.auto_session().close()
        rclpy.shutdown()



"""
Verbose example showing how to integrate Zenoh with asyncio_for_robotics.

This script demonstrates:
- Initializing Zenoh objects and creating a publisher with `auto_session()`.
- Publishing data in the background while running subscriber examples.
- Call to `python_discussion.discuss` then illustrate different asyncio-based
  subscription methods (`wait_for_value`, `wait_for_new`, `wait_for_next`,
  `listen`, and `listen_reliable`). Refere to `python_discussion.py` for details
- Properly shutting down Zenoh sessions to allow clean program exit.
"""
import asyncio
from contextlib import suppress

import zenoh

from asyncio_for_robotics.zenoh.session import auto_session
from asyncio_for_robotics.zenoh.sub import Sub
from asyncio_for_robotics.core._logger import setup_logger

from .python_discussion import brint, discuss
setup_logger("./")


async def talking_loop():
    brint(
        f"""asyncio_for_robotics does not provide zenoh publishers, it is out
          of scope. You remain in full control of publishing. You can get the
          zenoh session to declare a publisher by using `auto_session()`"""
    )
    pub = auto_session().declare_publisher("example/discussion")
    print(f"Zenoh started publishing onto {pub.key_expr}")
    try:
        count = 0
        while 1:
            pub.put(f"[Hello world! timestamp: {count/10:.1f}s]")
            count += 1
            await asyncio.sleep(0.1)
    finally:
        print(f"Zenoh stopped publishing onto {pub.key_expr}")
        pub.undeclare()


def get_str_from_msg(msg: zenoh.Sample):
    return msg.payload.to_string()


async def main():
    background_talker_task = asyncio.create_task(talking_loop())
    sub = Sub("example/**")
    try:
        await discuss(sub, get_str_from_msg)
    finally:
        sub.close()
        background_talker_task.cancel()


if __name__ == "__main__":
    brint(
        f"""
        Zenoh requires a zenoh session runing in the background.
        asyncio_for_robotics will automatically call `auto_session` if a
        zenoh session is not provided. This auto_session is a singleton
        (one per python process).
        """
    )
    asyncio.run(main())
    print()
    brint(
        f"""
    To finish and let python exit, the zenoh session needs to be closed. We can
    retrieve and close the one automatically created with `auto_session().close()`
    """
    )
    with suppress(zenoh.ZError): # why error here???
        auto_session().close()



import pytest

pytest.importorskip("rclpy")

import logging
from typing import Any, AsyncGenerator, Callable, Generator, Optional

import rclpy
from base_tests import (
    test_freshness,
    test_listen_one_by_one,
    test_listen_too_fast,
    test_reliable_extremely_fast,
    test_reliable_one_by_one,
    test_reliable_too_fast,
    test_wait_for_value,
    test_wait_new,
    test_wait_next,
)
from rclpy.qos import QoSProfile
from std_msgs.msg import String

import asyncio_for_robotics.ros2 as aros
from asyncio_for_robotics.core import BaseSub
from asyncio_for_robotics.core._logger import setup_logger
from asyncio_for_robotics.ros2.session import ThreadedSession, set_auto_session

setup_logger(debug_path="tests")
logger = logging.getLogger("asyncio_for_robotics.test")


@pytest.fixture(scope="module")
def session() -> Generator[aros.BaseSession, Any, Any]:
    logger.info("Starting rclpy and session")
    rclpy.init()
    set_auto_session(ThreadedSession())
    ses = aros.auto_session()
    yield ses
    logger.info("closing rclpy and session")
    ses.close()
    rclpy.shutdown()


topic = aros.TopicInfo(
    "test/something",
    String,
    QoSProfile(
        depth=10000,
    ),
)
TOPIC = topic


class SubProcessed(BaseSub[str]):
    def __init__(
        self,
        msg_type: type[String],
        topic: str,
        qos_profile: QoSProfile = aros.QOS_DEFAULT,
        session: Optional[aros.BaseSession] = None,
    ) -> None:
        super().__init__()
        self.ros_sub = aros.Sub(msg_type, topic, qos_profile, session)
        self.ros_sub.asap_callback.append(lambda x: self._input_data_asyncio(x.data))

    def close(self):
        self.ros_sub.close()


@pytest.fixture(scope="module")
def pub(session: aros.BaseSession) -> Generator[Callable[[str], None], Any, Any]:
    with session.lock() as node:
        publisher = node.create_publisher(*TOPIC.as_arg())

    def write_in_proc(input: str) -> None:
        publisher.publish(String(data=input))

    yield write_in_proc
    with session.lock() as node:
        node.destroy_publisher(publisher)


@pytest.fixture
async def sub(session) -> AsyncGenerator[BaseSub[str], Any]:
    s: BaseSub[str] = SubProcessed(*TOPIC.as_arg())
    yield s
    s.close()



import asyncio
import logging
from asyncio import AbstractEventLoop, Future
from typing import Optional

from rclpy.task import Future as RosFuture

logger = logging.getLogger(__name__)


def asyncify_future(
    ros_future: RosFuture,
    event_loop: Optional[AbstractEventLoop] = None,
) -> Future:
    """
    Convert a ROS Future into an asyncio Future.

    The asyncio Future will complete when the ROS Future completes,
    propagating its result, cancellation or exception.

    Args:
        ros_future: ROS Future.
        event_loop: Asyncio event loop to use. If None, uses asyncio.get_event_loop().

    Returns:
        Asyncio Future reflecting the ROS Future.
    """
    ao_future: Future = Future()
    if event_loop == None:
        event_loop = asyncio.get_event_loop()

    def ros_cbk(fut: RosFuture):
        nonlocal ao_future
        if fut.cancelled():
            event_loop.call_soon_threadsafe(ao_future.cancel)
            return
        if fut.done():
            exc = fut.exception()
            if exc is not None:
                event_loop.call_soon_threadsafe(ao_future.set_exception, exc)
            else:
                res = fut.result()  # type: ignore
                event_loop.call_soon_threadsafe(ao_future.set_result, res)

    # lock not necessary, ros seems safe
    ros_future.add_done_callback(ros_cbk)
    return ao_future



from .session import (
    set_auto_session,
    auto_session,
)
from .sub import Sub
from ..core.utils import soft_timeout, soft_wait_for, Rate

__all__ = [
    "soft_wait_for",
    "soft_timeout",
    "Rate",
    "auto_session",
    "set_auto_session",
    "Sub",
]



import pytest

pytest.importorskip("zenoh")
import asyncio
import logging
from contextlib import suppress
from typing import Any, AsyncGenerator, Callable, Generator, Optional, Union

import zenoh
from base_tests import (
    test_freshness,
    test_listen_one_by_one,
    test_listen_too_fast,
    test_reliable_extremely_fast,
    test_reliable_one_by_one,
    test_reliable_too_fast,
    test_wait_for_value,
    test_wait_new,
    test_wait_next,
)

from asyncio_for_robotics.core import BaseSub
from asyncio_for_robotics.zenoh import (
    Sub,
    auto_session,
    set_auto_session,
    soft_timeout,
    soft_wait_for,
)

logger = logging.getLogger("asyncio_for_robotics.test")


@pytest.fixture(scope="module", autouse=True)
def session() -> Generator[zenoh.Session, Any, Any]:
    set_auto_session(zenoh.open(zenoh.Config()))
    ses = auto_session()
    yield ses
    if not auto_session().is_closed():
        with suppress(zenoh.ZError):
            ses.close()


@pytest.fixture
def pub(session) -> Generator[Callable[[str], None], Any, Any]:
    pub_topic = "test/something"
    logger.debug("Creating PUB-%s", pub_topic)
    p: zenoh.Publisher = auto_session().declare_publisher(
        pub_topic, reliability=zenoh.Reliability.RELIABLE
    )

    def pub_func(input: str):
        p.put(input.encode())

    yield pub_func
    if not auto_session().is_closed():
        logger.debug("closing PUB-%s", pub_topic)
        p.undeclare()


class SubProcessed(BaseSub[str]):
    def __init__(
        self,
        key_expr: Union[zenoh.KeyExpr, str],
        session: Optional[zenoh.Session] = None,
    ) -> None:
        super().__init__()
        self.ros_sub = Sub(key_expr, session)
        self.ros_sub.asap_callback.append(self._preprocess_cbk)

    def _preprocess_cbk(self, payload: zenoh.Sample):
        self._input_data_asyncio(payload.payload.to_string())

    def close(self):
        self.ros_sub.close()


@pytest.fixture
async def sub(session) -> AsyncGenerator[BaseSub[str], Any]:
    s: BaseSub[str] = SubProcessed("test/**")
    yield s
    s.close()


# async def test_wait_for_value(pub: zenoh.Publisher, sub: Sub):
#     payload = b"hello"
#     pub.put(payload)
#     sample = await soft_wait_for(sub.wait_for_value(), 1)
#     assert not isinstance(sample, TimeoutError), f"Did not receive response in time"
#     assert sample.payload.to_bytes() == payload
#     assert str(sample.key_expr) == str(pub.key_expr)
#
#
# async def test_wait_new(pub: zenoh.Publisher, sub: Sub):
#     payload = b"hello"
#     pub.put(payload)
#     sample = await sub.wait_for_value()
#     assert not isinstance(sample, TimeoutError), f"Should get a message"
#
#     wait_task = sub.wait_for_new()
#     new_sample = await soft_wait_for(wait_task, 0.1)
#     assert isinstance(new_sample, TimeoutError), f"Should not get a message"
#
#     wait_task = sub.wait_for_new()
#     pub.put(payload)
#     new_sample = await soft_wait_for(wait_task, 0.1)
#     assert not isinstance(new_sample, TimeoutError), f"Should get the message"
#     assert new_sample.payload.to_bytes() == payload
#
#
# async def test_wait_next(pub: zenoh.Publisher, sub: Sub):
#     first_payload = b"hello"
#     pub.put(first_payload)
#     sample = await sub.wait_for_value()
#     assert not isinstance(sample, TimeoutError), f"Should get a message"
#
#     wait_task = sub.wait_for_next()
#     new_sample = await soft_wait_for(wait_task, 0.1)
#     assert isinstance(new_sample, TimeoutError), f"Should not get a message"
#
#     wait_task = sub.wait_for_next()
#     pub.put(first_payload)
#     for other_payload in range(10):
#         await asyncio.sleep(0.001)
#         pub.put(str(other_payload))
#
#     new_sample = await soft_wait_for(wait_task, 0.1)
#     assert not isinstance(new_sample, TimeoutError), f"Should get the message"
#     assert new_sample.payload.to_bytes() == first_payload
#
#
# async def test_listen_one_by_one(pub: zenoh.Publisher, sub: Sub):
#     last_payload = "hello"
#     pub.put(last_payload)
#     sample_count = 0
#     put_count = 1
#     max_iter = 20
#     async for sample in sub.listen():
#         sample_count += 1
#         assert sample.payload.to_string() == last_payload
#         if sample_count >= max_iter:
#             break
#         last_payload = f"hello{sample_count}"
#         pub.put(last_payload)
#         put_count += 1
#
#     assert put_count == sample_count == max_iter
#
#
# async def test_listen_too_fast(pub: zenoh.Publisher, sub: Sub):
#     last_payload = "hello"
#     pub.put(last_payload)
#     pub.put(last_payload)
#     sample_count = 0
#     put_count = 2
#     max_iter = 20
#     await asyncio.sleep(0.001)
#     async for sample in sub.listen():
#         sample_count += 1
#         assert sample.payload.to_string() == last_payload
#         if sample_count >= max_iter:
#             break
#         last_payload = f"hello{sample_count}"
#         pub.put(last_payload)
#         put_count += 1
#         await asyncio.sleep(0.001)
#         last_payload = f"hello{sample_count}"
#         pub.put(last_payload)
#         put_count += 1
#         await asyncio.sleep(0.001)
#
#     assert put_count / 2 == sample_count == max_iter
#
#
# async def test_reliable_one_by_one(pub: zenoh.Publisher, sub: Sub):
#     last_payload = "hello"
#     pub.put(last_payload)
#     sample_count = 0
#     put_count = 1
#     max_iter = 20
#     async for sample in sub.listen_reliable():
#         sample_count += 1
#         assert sample.payload.to_string() == last_payload
#         if sample_count >= max_iter:
#             break
#         last_payload = f"hello{sample_count}"
#         pub.put(last_payload)
#         put_count += 1
#
#     assert put_count == sample_count == max_iter
#
#
# async def test_reliable_too_fast(pub: zenoh.Publisher, sub: Sub):
#     data = list(range(30))
#     put_queue = [str(v) for v in data]
#     put_queue.reverse()
#     received_buf = []
#     listener = sub.listen_reliable(fresh=True, queue_size=len(data))
#     await asyncio.sleep(0.1)
#     pub.put(put_queue.pop())
#     await asyncio.sleep(0.001)
#     async with soft_timeout(2):
#         async for sample in listener:
#             payload = int(sample.payload.to_string())
#             received_buf.append(payload)
#             if len(received_buf) >= len(data):
#                 break
#             if put_queue != []:
#                 pub.put(put_queue.pop())
#                 await asyncio.sleep(0.001)
#             if put_queue != []:
#                 pub.put(put_queue.pop())
#                 await asyncio.sleep(0.001)
#
#     assert set(data) == set(received_buf)
#     assert data == received_buf
#
#
# async def test_reliable_extremely_fast(pub: zenoh.Publisher, sub: Sub):
#     data = list(range(30))
#     put_queue = [str(v) for v in data]
#     put_queue.reverse()
#     received_buf = []
#     listener = sub.listen_reliable(fresh=True, queue_size=len(data))
#     pub.put(put_queue.pop())
#     async with soft_timeout(2):
#         async for sample in listener:
#             payload = int(sample.payload.to_string())
#             received_buf.append(payload)
#             if len(received_buf) >= len(data):
#                 break
#             if put_queue != []:
#                 pub.put(put_queue.pop())
#             if put_queue != []:
#                 pub.put(put_queue.pop())
#
#     assert set(data) == set(received_buf)
#
#
# async def test_freshness(pub: zenoh.Publisher, sub: Sub):
#     payload = "hello"
#
#     new = sub.wait_for_new()
#     pub.put(payload)
#     await new
#     sample = await soft_wait_for(anext(sub.listen(fresh=False)), 0.1)
#     assert not isinstance(sample, TimeoutError), f"Should get the message"
#     assert sample.payload.to_string() == payload
#
#     new = sub.wait_for_new()
#     pub.put(payload)
#     await new
#     sample = await soft_wait_for(anext(sub.listen_reliable(fresh=False)), 0.1)
#     assert not isinstance(sample, TimeoutError), f"Should get the message"
#     assert sample.payload.to_string() == payload
#     await sub.wait_for_value()
#
#     new = sub.wait_for_new()
#     pub.put(payload)
#     await new
#     sample = await soft_wait_for(anext(sub.listen(fresh=True)), 0.1)
#     assert isinstance(sample, TimeoutError), f"Should NOT get the message"
#
#     new = sub.wait_for_new()
#     pub.put(payload)
#     await new
#     sample = await soft_wait_for(anext(sub.listen_reliable(fresh=True)), 0.1)
#     assert isinstance(sample, TimeoutError), f"Should NOT get the message"



from dataclasses import dataclass, field
from typing import Any, Dict, Final, Generic, NamedTuple, Tuple, TypeVar

from rclpy.qos import (
    DurabilityPolicy,
    HistoryPolicy,
    QoSProfile,
    ReliabilityPolicy,
    qos_profile_system_default,
)

#: Default qos
QOS_DEFAULT: Final = qos_profile_system_default

#: "always available" qos
QOS_TRANSIENT: Final = QoSProfile(
    reliability=ReliabilityPolicy.RELIABLE,
    history=HistoryPolicy.KEEP_LAST,
    depth=10,
    durability=DurabilityPolicy.TRANSIENT_LOCAL,
)


_MsgType = TypeVar("_MsgType")


@dataclass(frozen=True, slots=True)
class TopicInfo(Generic[_MsgType]):
    """Precisely describes a ROS2 topic

    Attributes:
        name:
        msg_type:
        qos:
    """

    topic: str
    msg_type: _MsgType
    qos: QoSProfile = field(default_factory=lambda *_, **__: QOS_DEFAULT)

    def as_arg(self) -> Tuple[_MsgType, str, QoSProfile]:
        return (self.msg_type, self.topic, self.qos)

    def as_kwarg(self) -> Dict[str, Any]:
        return {"msg_type": self.msg_type, "topic": self.topic, "qos_profile": self.qos}



import asyncio
import logging
from asyncio import AbstractEventLoop, Future
from typing import Generic, Optional, Protocol, TypeVar

from rclpy.client import Client as RosClient
from rclpy.impl.implementation_singleton import rclpy_implementation as _rclpy
from rclpy.qos import QoSProfile
from rclpy.service import Service as RosService
from rclpy.task import Future as RosFuture

from ..core.sub import BaseSub
from .future import asyncify_future
from .session import BaseSession, auto_session
from .utils import QOS_DEFAULT, TopicInfo

logger = logging.getLogger(__name__)

_ReqT = TypeVar("_ReqT")
_ResT = TypeVar("_ResT")


# Generic base class for a service definition
class ServiceType(Protocol[_ReqT, _ResT]):
    Request: type[_ReqT]
    Response: type[_ResT]


_SrvType = ServiceType


class Responder(Generic[_ReqT, _ResT]):
    def __init__(
        self,
        request: _ReqT,
        response: _ResT,
        srv: RosService,
        event_loop: AbstractEventLoop,
    ) -> None:
        """Sends a response to a request.

        Use:
          - Get request data: `the_request = responder.request`
          - Set the response: `responder.response = my_response`
          - Send the response: `responder.send()`

        .. Important:
            User should not instanciate this object.

        .. Important:
            It is possible to send several replies. Do not.
        """
        self.request: _ReqT = request
        self.response: _ResT = response
        self._srv: RosService = srv
        self._event_loop = event_loop
        self._header_ready: Future = Future(loop=event_loop)

    def _set_header(self, header) -> None:
        logger.debug(f"Header set in respondable obj")
        self._header_ready.set_result(header)

    def send(self, response: Optional[_ResT] = None) -> None:
        """Sends the response on the transport (rmw)

        This reimplement the `Service.send_response` of ros, because we overode
        it previously with `response_overide`.
        """
        if response is None:
            response = self.response
        logger.debug("Properly replying to request")
        if not self._header_ready.done():
            raise AttributeError(f"Header is missing! something went terribly wrong")
        header = self._header_ready.result()
        if not isinstance(
            response,
            self._srv.srv_type.Response,  # type: ignore
        ):
            raise TypeError(
                f"Response is of the wrong type: \n  - {type(response)=}\n  - {self._srv.srv_type.Response=}"
            )
        with self._srv.handle:
            c_implementation = self._srv._Service__service  # type: ignore
            if isinstance(header, _rclpy.rmw_service_info_t):
                c_implementation.service_send_response(response, header.request_id)
            elif isinstance(header, _rclpy.rmw_request_id_t):
                c_implementation.service_send_response(response, header)
            else:
                raise TypeError(f"Header is of the wrong type: {type(header)=}")


def response_overide(response: Responder, header) -> None:
    """Replaces the callback behavior of ros.

    By default when service callback is triggered, the response is returned,
    then passed to `Service.send_response` that then sends it on the tranport
    (rmw). However, we don't want to respond yet, we wanna respond later. So we
    overide Service.send_response with nothing (this function)...

    Not exactly nothing. `Service.send_response` has the critical "header"
    information that we need to save. This header is saved in the Responder
    instance later used when the user sends the response.
    """
    responder = response
    logger.debug("Response and header intercepted")
    responder._event_loop.call_soon_threadsafe(responder._set_header, header)


class Server(BaseSub[Responder[_ReqT, _ResT]]):
    def __init__(
        self,
        msg_type: type[_SrvType[_ReqT, _ResT]],
        topic: str,
        qos_profile: QoSProfile = QOS_DEFAULT,
        session: Optional[BaseSession] = None,
    ) -> None:
        """Implements an async ROS 2 service server.

        This is a standard `afor` subscriber, where the data stream is made of
        `Responder` objects. The `Responder` holds the request and response.
        The response is then send using `Responder.send()`

        Args:
            msg_type:
            topic:
            qos_profile:
            session:
        """
        self.session: BaseSession = self._resolve_session(session)
        self.topic_info = TopicInfo(topic=topic, msg_type=msg_type, qos=qos_profile)
        self.srv = self._resolve_sub(self.topic_info)
        super().__init__()

    @property
    def name(self) -> str:
        try:
            return f"ROS2-SRV-{self.srv.srv_name}"
        except:
            return f"ROS2-SRV-{self.topic_info.topic}"

    def _resolve_session(self, session: Optional[BaseSession]) -> BaseSession:
        return auto_session(session)

    def _resolve_sub(self, topic_info: TopicInfo) -> RosService:
        logger.debug("%s requesting lock for creation", self.name)
        with self.session.lock() as node:
            serv_modified = node.create_service(
                srv_type=topic_info.msg_type,
                srv_name=topic_info.topic,
                qos_profile=topic_info.qos,
                callback=self._incomming_request_cbk,
            )
            ### VVV IMPORTANT VVV ###
            ###                   ###
            serv_modified.send_response = response_overide
            ###                   ###
            ### ^^^           ^^^ ###
        return serv_modified

    def _incomming_request_cbk(self, request: _ReqT, response: _ResT) -> Responder:
        """Fake service callback

        actually returning the responder to be later intercepted, and the
        header to be later set.
        """
        responder_for_user: Responder[_ReqT, _ResT] = Responder(
            request, response, self.srv, self._event_loop
        )

        def execute_in_asyncio_thread():
            responder_for_user._header_ready.add_done_callback(
                lambda *_: self._differed_header_ready_cbk(responder_for_user)
            )

        self._event_loop.call_soon_threadsafe(execute_in_asyncio_thread)
        return responder_for_user

    def _differed_header_ready_cbk(self, responder_for_user: Responder):
        """Executes once the header is set.

        Once the header of the request is set, the responder is fully usable.
        This needs to be differed in a cbk because it happens after the return
        statement of _incomming_request_cbk
        """
        try:
            healty = self.input_data(responder_for_user)
            if not healty:
                self.session._node.destroy_service(self.srv)
        except Exception as e:
            logger.error(e)

    def close(self):
        with self.session.lock() as node:
            if not node.executor.context.ok():
                return
            node.destroy_service(self.srv)


class Client(Generic[_ReqT, _ResT]):
    def __init__(
        self,
        msg_type: type[_SrvType[_ReqT, _ResT]],
        topic: str,
        qos_profile: QoSProfile = QOS_DEFAULT,
        session: Optional[BaseSession] = None,
    ) -> None:
        """Implements an async ROS 2 service client.

        The `Client.call(...)` method will send a request and return its
        asyncio.Future response.
        """
        self.session: BaseSession = self._resolve_session(session)
        self._event_loop = asyncio.get_event_loop()
        self.topic_info = TopicInfo(topic=topic, msg_type=msg_type, qos=qos_profile)
        self.cli: RosClient = self._resolve_sub(self.topic_info)

    def _resolve_session(self, session: Optional[BaseSession]) -> BaseSession:
        return auto_session(session)

    def _resolve_sub(self, topic_info: TopicInfo) -> RosClient:
        logger.debug("%s requesting lock for creation", self.name)
        with self.session.lock() as node:
            client = node.create_client(
                srv_type=topic_info.msg_type,
                srv_name=topic_info.topic,
                qos_profile=topic_info.qos,
            )
        return client

    async def wait_for_service(self, polling_rate: float = 0.25):
        """
        Wait for a service server to become ready.

        .. Note:
            By default in ROS 2 this is a busy wait in a while loop. Crazy unga
            bunga. Me too unga bunga.

        Args:
            polling_rate: Rate (in s) at which to check for readiness.

        Returns:
            As soon as a server becomes ready.
        """
        logger.debug("%s waiting for server", self.name)
        while 1:
            if self.cli.service_is_ready():
                logger.debug("%s server is here", self.name)
                return
            else:
                await asyncio.sleep(polling_rate)

    def call(self, req: _ReqT) -> Future[_ResT]:
        """Calls the service and returns the response as asyncio.Future

        Args:
            req: Request to send to the server.

        Returns:
            response as an asyncio.Future
        """
        logger.debug("%s making request", self.name)

        def dbg(ros_response):
            logger.debug("%s got response", self.name)

        # lock not necessary, ros seems safe
        ros_fut: RosFuture = self.cli.call_async(req)
        future: Future[_ResT] = asyncify_future(ros_fut, self._event_loop)
        future.add_done_callback(dbg)
        return future

    @property
    def name(self) -> str:
        try:
            return f"ROS2-CLI-{self.cli.srv_name}"
        except:
            return f"ROS2-CLI-{self.topic_info.topic}"

    def close(self):
        with self.session.lock() as node:
            if not node.executor.context.ok():
                return
            node.destroy_client(self.cli)



"""
Example ROS 2 subscriber using asyncio integration.

This example demonstrates how to:
- Subscribe to multiple ROS 2 topics concurrently with asyncio.
- React immediately to the first new message from any topic.
- Use `soft_timeout` to wait briefly for additional messages while still
  resuming as soon as all subscriptions have received data.
- Distinguish between new and previously received messages.
- Safely retrieve the most recent values even if one topic has never
  produced data.

Run this script using `python3 -m asyncio_for_robotics.example.ros2_double_listener`
Run this script side by side with `python3 -m asyncio_for_robotics.example.ros2_double_talker`
"""

import asyncio
from contextlib import suppress

import rclpy
from colorama import Fore  # colorama is included with ros so not dependency
from std_msgs.msg import String

from asyncio_for_robotics import soft_timeout
from asyncio_for_robotics.ros2 import Sub, TopicInfo, auto_session

TOPIC1 = TopicInfo(msg_type=String, topic="example/talker1")
TOPIC2 = TopicInfo(msg_type=String, topic="example/talker2")


async def main():
    sub1 = Sub(TOPIC1.msg_type, TOPIC1.topic, TOPIC1.qos)
    sub2 = Sub(TOPIC2.msg_type, TOPIC2.topic, TOPIC2.qos)
    while 1:
        task1 = asyncio.create_task(sub1.wait_for_new())
        task2 = asyncio.create_task(sub2.wait_for_new())

        # waits for a message on any of the topics
        await asyncio.wait([task1, task2], return_when=asyncio.FIRST_COMPLETED)

        # this waits for an additional 0.2s after receiving the first message
        # However it continues immediately once all messages are received!
        async with soft_timeout(0.2):
            await asyncio.wait([task1, task2], return_when=asyncio.ALL_COMPLETED)

        # gets the latest data if it exists
        # necessary because one of the topic might get a message, while the
        # other never received anything
        sub1_has_data = sub1.alive.is_set()
        sub2_has_data = sub2.alive.is_set()
        data1 = (await sub1.wait_for_value()).data if sub1_has_data else None
        data2 = (await sub2.wait_for_value()).data if sub2_has_data else None
        # adds fancy colors depending if data is new or old
        freshness1 = (
            f"{Fore.GREEN}new{Fore.RESET}"
            if task1.done()
            else f"{Fore.RED}old{Fore.RESET}"
        )
        freshness2 = (
            f"{Fore.GREEN}new{Fore.RESET}"
            if task2.done()
            else f"{Fore.RED}old{Fore.RESET}"
        )
        print(
            f"""
Latest received messages:
  - Pub #1: ({freshness1}) {data1}
  - Pub #2: ({freshness2}) {data2}
              """
        )


if __name__ == "__main__":
    rclpy.init()
    try:
        # suppress, just so we don't flood the terminal on exit
        with suppress(KeyboardInterrupt, asyncio.CancelledError):
            asyncio.run(main())  # starts asyncio executor
    finally:
        # cleanup. `finally` statment always executes
        auto_session().close()
        rclpy.shutdown()
